# -*- coding: utf8 -*-
from __future__ import with_statement # This isn't required in Python 2.6
import struct
import sys
import os
from urllib2 import urlopen, HTTPError
import socket
import subprocess as sp
import logging

__author__ = "Gary Dusbabek gdusbabek@gmail.com"
__version__ = "0.1"
__copyright__ = "None. 2009 Gary Dusbabek"
__license__ = "None. I release this code into the public domain. Have at it."

"""
This is a simple generic ID3 parser that attempts to extract text tags from
mp3 files.  All strings returned are unicode, so no fretting there.  I also
took pains to create a way to capture the mp3 image, if there is one. To disable,
set 'pic_saver' to None.  There is a slight dependency on the 'convert' utility
supplied by Imagemagick to save APIC tags.

Usage:

id3 = PathId3(<path to a file>)
-or-
id3 = StreamId3(<stream of some sort--something that can be read>)
getRemoteTags(url) wraps this approach.

Then everything you want is in the id3.tags dictionary.  I'm discarding everything
but artist, ablum, title, year and track.  However, it would be simple to
throw every text frame into the dict.
"""

logger = logging.getLogger("id3")
logger.setLevel(logging.DEBUG)
# you can add handlers if they want from their code.

ID3v2_2 = "Id3v2.2"
ID3v2_3 = "Id3v2.3"
ID3v2_4 = "Id3v2.4"
INVALID = "Invalid"

MAX_TAG_LENGTH = 5000000
# frame header signatures for the various id3 versions.
V2_VERSION = struct.pack("BB", 2, 0)
V3_VERSION = struct.pack("BB", 3, 0)
V4_VERSION = struct.pack("BB", 4, 0)

# useful to detect 0x00 and 0x00,0x00.
NULL = struct.pack("B", 0)
NULLNULL = struct.pack("BB", 0, 0)

FF = struct.pack("B", 0xFF)

# needed to detect unicode.
BOM1 = struct.pack("BB", 255, 254)
BOM2 = struct.pack("BB", 254, 255)

# I couldn't get urllib2 to behave responsibly when a website would take too
# long to respond. Teh internets pointed me towards this ugly workaround.
timeout = 10
socket.setdefaulttimeout(10)

# example pic saver.
def picSaver(d):
    """ saves picture data to a specified directory. """
    mime = d["mime"]
    data = d["data"]
    artist = d["artist"]
    album = d["album"]
    title = d["title"]
    ext = mime.split("/")[1];
    fname = os.path.join("/home/garyd/dusbabek_svn/mp3_blog_scraper/pic", str(hash(artist+album+title))+"."+ext)
    f = open(fname, "wb")
    f.write(data)
    f.close()
    # now resample it to be 75px.
    sp.call(["convert", "-size", "75", fname, fname])

#pic_saver = picSaver
pic_saver = None

ID3_FIELDS = ["album", "artist", "track", "title", "year", "picture"]

def extract_binary_tag(instream, path):
    """Extracts id3 tag information and saves it, so that I have binary data to ship with my unit tests.
    A glaring precondtion for this method is that you know what you're doing and that the stream in question
    actually refers to real ID3 data.  If not, you're screwed."""
    outstream = file(path, "wb")
    outstream.write(instream.read(3)) # ID3
    version = instream.read(2)
    outstream.write(version)
    outstream.write(instream.read(1)) # header byte
    packed_header_len = instream.read(4)
    outstream.write(packed_header_len)
    real_header_len = Helpers.compute_tag_length(struct.unpack("BBBB", packed_header_len))
    if real_header_len > MAX_TAG_LENGTH:
        raise Mp3Exception("My, what big tags you have, grandma!")
    outstream.write(instream.read(real_header_len))
    outstream.close()
    instream.close()

class StreamId3:
    """Reads id3 information from a stream like object (opened file, opened url, etc.)"""
    def __init__(self, stream, path):
        fileIdentifier = stream.read(3)
        # check identifier to see if tag is present.
        if fileIdentifier != "ID3":
            # no tags. give it something empty.
            logger.debug("No tags")
            self._tags = BasicId3(stream, path)
            self.version = INVALID
        else:
            # next two bytes contain the ID3v2.x versin $03,$00 or $04,$00
            version = stream.read(2)
            if version == V2_VERSION:
                logger.debug("Id3 2.2")
                self._tags = Id3v22(stream, path)
                self.version = ID3v2_2
            elif version == V3_VERSION:
                logger.debug("Id3 2.3")
                self._tags = Id3v23(stream, path)
                self.version = ID3v2_3
            elif version == V4_VERSION:
                logger.debug("Id3 2.4")
                self._tags = Id3v24(stream, path)
                self.version = ID3v2_4
            else:
                logger.debug("Unidentified ID3 version: " + path)

    def __getattr__(self, name):
        if name in ID3_FIELDS:
            return getattr(self.__dict__["_tags"], name)
        else:
            return self.__dict__[name]


class PathId3(StreamId3):
    """ Used to create an id3 from a filesystem path. """
    def __init__(self, path):
        StreamId3.__init__(self, open(path, 'rb'), path)

## 2.2
def simple(data):
    """ generic handler. echoes data. 2.2 text frames use this. """
    return data

## 2.3, 2.4
def text_frame_with_enc(data):
    """ return 2.3/4 text frames as unicode. """
    # 1==unicode, 2==iso8859-1, except BOM is always unicode
    enc = struct.unpack("B", data[0:1])[0]
    if len(data) > 3:
        bom = data[1:3]
    else:
        bom = None
    if bom == BOM1 or bom == BOM2:
        return unicode(data[3:], "utf16").rstrip()
    elif enc == 1:
        return unicode(unicode(data[1:], "utf16").rstrip()).rstrip()
    else:
        # last ditch effort: make every attempt to return good data.
        # do note that iso8859-1 is a perfectly valid encoding as per the spec.
        for enc in ["utf8", "iso8859-1", "cp1252"]:
            try:
                string = unicode(data[1:], enc).rstrip()
                logger.debug("ENC " + enc + " " + str(len(string)))
                # I really don't know what's going on here. I came across an id3
                # encoding in iso-8859-1, listing a tag length as 10, but only
                # having 9 characters of data and then a null. rstrip didn't seem
                # to be removing the null, so I manually do it here.
                while len(string) > 0 and ord(string[len(string)-1]) is 0:
                    string = string[0:len(string)-1]
                return string
            except UnicodeDecodeError:
                pass

        # I give up.
        return unicode(data[1:], errors="ignore").rstrip()

## generic
def noimpl(data):
    return ""

def pic(data):
    """ get the picture data out of a 2.2 frame."""
    sz = len(data)
    pos = 0
    enc = struct.unpack("B", data[pos])[0]
    logger.debug("PIC ENC " + str(enc))
    pos += 1
    mime = data[pos:pos+3]
    pos += 3;
    type = struct.unpack("B", data[pos])[0]
    pos += 1 # get past the type
    desc = ""
    hasDesc = False
    while data[pos] != NULL:
        hasDesc = True
        desc += data[pos]
        pos += 1;
    desc = unicode(desc, "utf8")
    logger.debug("PIC DESC " + desc)
    pos += 1 # get past the null.
    while hasDesc and data[pos] is NULL and pos < sz:
        pos += 1
    return {"mime" : mime, "data": data[pos:]}

def apic(data):
    """get the picture data out of 2.3 and 2.4 frames."""
    sz = len(data)
    pos = 0
    enc = struct.unpack("B", data[pos])[0]
    logger.debug("APIC ENC " + str(enc))
    pos += 1
    mime = ""
    while data[pos] != NULL:
        mime += data[pos]
        pos += 1
    logger.debug("APIC MIME " + mime)
    pos += 1 # get past the null
    type = struct.unpack("B", data[pos])[0]
    logger.debug("APIC TYPE " + str(type))
    pos += 1 # get past the type
    desc = ""

    # here is a dilemma: the descrption may be omitted entirely. the only approach
    # that can be taken is to read until a null/end is reached and try to create
    # some text.  If that doesn't work, back up an assume we've encroached on
    # the image portion.  Whacky, but I don't see a way around it, and I have
    # come across it in the wild (looks to have come from WMP).

    # or I could just say to hell with it and bail on extracting the image
    # for this particular file.  
    
    hasDesc = False

    if data[pos] != FF:
        while data[pos] != NULL:
            hasDesc = True
            desc += data[pos]
            pos += 1;
        for encoding in ["utf8", "iso8859-1", "cp1252"]:
            try:
                desc = unicode(desc, encoding)
                break
            except UnicodeDecodeError:
                pass

        logger.debug("APIC DESC " + desc)
        pos += 1 # get past the null.
    else:
        logger.debug("APIC has nonexisting description. bastards.")
        
    while hasDesc and data[pos] is NULL and pos < sz:
        pos += 1
    return {"mime" : mime, "data": data[pos:]}

# special frame handlers for non-text frames. You can extend these by adding
# your own key:func pairs.
_HANDLERSv23 = {"APIC":apic}
_HANDLERSv22 = {"PIC":pic}

class BasicId3:
    """Parsing is more or less the same regardless of id3 version.  The only
    thing that really changes are some field lengths (customized in child classes)."""
    def __init__(self, file, path):
        self.path = path
        self.album = ""
        self.artist = ""
        self.track = ""
        self.title = ""
        self.year = ""
        self.picture = None

    def process(self, file):
        """ This is where the work is done. It is a generic method that reads frames and passes them
        off to handler functions. These handlers are determined in subclasses."""
        # 5 bytes have already been read
        pic = None
        isSynchronized, isExtended, isExperimental, footer = Helpers.grok_header_flag(struct.unpack("B", file.read(1))[0])
        tagSize = Helpers.compute_tag_length(struct.unpack("BBBB", file.read(4)))
        if tagSize > MAX_TAG_LENGTH: return # low memory system.
        logger.debug("Tag size " + str(tagSize))
        read = 0
        while (read < tagSize):
            frameId = file.read(self.FRAME_ID_LEN).strip()
            read += self.FRAME_ID_LEN
            if frameId == self.ZEROS:
                break
            if len(frameId) < self.FRAME_ID_LEN: break
            # if we read a sync signal, we've gone too far.
            if Helpers.contains_mp3_sync_signal(frameId): break

            try:
                frameSz = Helpers.make_int(struct.unpack(self.FRAME_SZ_FMT, file.read(len(self.FRAME_SZ_FMT))), 8)
                # if the frame size is larger than the tag size, we have a problem.
                # this is unrecoverable (unless there is bug in make_int()), as we don't know where the next frame might
                # start. Give up.
                if frameSz > tagSize:
                    raise Mp3Exception("Reported frame size is larger than reported tag size. %d, %d %s" % (frameSz, tagSize, frameId))
                read += len(self.FRAME_SZ_FMT)
                flags, flagsLen = self.getFlags(file)
                read += flagsLen
                if frameSz == 0:
                    continue
                frameData = file.read(frameSz)
                read += frameSz
                if frameId in self.HANDLERS:
                    info = self.HANDLERS[frameId](frameData)
                    if self.HANDLERS[frameId] is apic:
                        # if there is a pic_saver defined, set pic. It gets saved
                        # at the end, after processing is finished.
                        if pic_saver:
                            pic = info
                        setattr(self, self.frameMap[frameId], info)
                elif frameId[0] == "T":
                    try:
                        info = self.TXT_HANDLER(frameData)
                    except UnicodeDecodeError:
                        # this is bad. log some details and bail.
                        logger.debug("Problem decoding frame %s %d" % (frameId, frameSz))
                        raise

                    logger.debug("%s is {%s}" % (frameId, info))
                else:
                    info = noimpl(frameData)
                if info and frameId in self.frameMap:
                    setattr(self, self.frameMap[frameId], info)
                else:
                    pass
#                    if verbose: logger.debug("Unused tag {%s} with value {%s}" % (frameId, info))
            except OverflowError:
                logger.debug("couldn't make int (probably misaligned data stream).")
                break;
            if tagSize - read < 10:
                # not enough space for another frame.grokHeaderFlag
                read = tagSize
        if pic:
            pic_saver(dict(zip(
                ["mime","data","artist","album","title"],
                [pic["mime"], pic["data"], self.artist, self.album, self.title])))

    def getFlags(self, file):
        return "", 0
    
class CrapId3(BasicId3):
    """A CrapId3 wraps a basic (empty) set of tags with an error message."""
    def __init__(self, errType, msg):
        BasicId3.__init__(self, None, None)
        self.errType = errType
        self.msg = msg

class Id3v22(BasicId3):
    """ id3 2.2 """
    def __init__(self, file, path):
        BasicId3.__init__(self, file, path)
        self.FRAME_ID_LEN = 3
        self.FRAME_SZ_FMT = "BBB"
        self.ZEROS = struct.pack("BBB", 0, 0, 0)
        self.HANDLERS = _HANDLERSv22
        self.TXT_HANDLER = text_frame_with_enc
        self.frameMap = {"TAL":"album", "TP1":"artist", "TRK":"track", "TT2":"title", "TYE":"year", "PIC":"picture"}
        self.process(file)

class Id3v23(BasicId3):
    """ id3 2.3 """
    def __init__(self, file, path):
        BasicId3.__init__(self, file, path)
        self.FRAME_ID_LEN = 4
        self.FRAME_SZ_FMT = "BBBB"
        self.ZEROS = struct.pack("BBBB", 0, 0, 0, 0)
        self.HANDLERS = _HANDLERSv23
        self.TXT_HANDLER = text_frame_with_enc
        self.frameMap = {"TALB":"album", "TPE1":"artist", "TRCK":"track", "TIT2":"title", "TYER":"year", "APIC":"picture"}
        self.process(file)

    def getFlags(self, file):
        return struct.unpack("BB", file.read(2)), 2

class Id3v24(Id3v23):
    """ id3 2.4 (more or less same as 2.3) """
    def __init__(self, file, path):
        BasicId3.__init__(self, file, path)

class Mp3Exception(Exception):
    pass

class Helpers:

    @classmethod
    def contains_mp3_sync_signal(cls, string):
        """ looks for an mp3 sync signal (not part of id3) """
        if struct.unpack("B", string[0:1])[0] & 0xff == 0xff:
            if struct.unpack("B", string[1:2])[0] & 0xe0 == 0xe0:
                return True
            else:
                return False
        else:
            return False

    @classmethod
    def make_int(cls, bytes, shift):
        """ makes an integer out of a byte array. shift indicates the amount of
        significant bits in each byte. id3 format can have 28 bit integers encoded in 32 bits. """
        res = 0;
        for byte in bytes:
            res <<= shift
            res |= byte
        return res

    @classmethod
    def grok_header_flag(cls, byte):
        """ 0:synchronized, 1:extended, 2:experimental, 3:footer(2.4 only) """
        sync = (byte & 0x80) > 0
        extended = (byte & 0x40) > 0
        experimental = (byte & 0x20) > 0
        footer = (byte & 0x10) > 0
        return sync, extended, experimental, footer

    @classmethod
    def compute_tag_length(cls, bytes):
        """ four bytes. msbit is always zero. convert to 28 bit unsigned. """
        return Helpers.make_int(bytes, 7)

def getRemoteTags(path):
    """ Utility method meant to be called from the outside. """
    try:
        mp3 = urlopen(path)
        id3 = StreamId3(mp3, path)
        return id3
    except:
        logger.error("%s %s" % (str(sys.exc_info()[0]), sys.exc_info()[1]))
        return CrapId3(str(sys.exc_info()[0]), sys.exc_info()[1])
    
